import sys
import networkx as nx
import random
import time

#k, r
num_bitmasks = 10
bit = 20

number_of_cascade = 1

#get edge list from a file
def create_graph(edgelist_file):
    g=nx.Graph()

    f=open(edgelist_file,'r')
    lines=f.readlines()
    f.close()
    for line in lines:
        #weight=eval(weight)
        temp=line.split()
        index1=int(temp[0])
        index2=int(temp[1])
        weight=0.6
        g.add_edge(index1,index2,weight=weight)

    return g

#get groups from a file
def get_groups(groups_file):

    f=open(groups_file,'r')
    lines=f.readlines()
    f.close()

    group_dict={}
    invert_group_dict={}
    for line in lines:
        temp=line.split(",")
        group_id = temp[0].strip()
        group_list = map( lambda x: int(x.strip()), temp[1:] )
        for node in temp[1:]:
            node_id = int(node.strip())
            invert_group_dict[node_id] = group_id

        group_dict.setdefault(group_id, group_list)

    return (group_dict, invert_group_dict)

#get timeseries from a file
def get_timeseries(infection_num_file):
    f=open(infection_num_file,'r')
    lines=f.readlines()
    f.close()
    timeseries={}
    max_t = int(  (lines[0].split(","))[1]  )
    for line in lines[1:]:
        temp=line.split(",")
        group_id= temp[0].strip()
        infection_list = [0] * max_t
        timeseries.setdefault(group_id, infection_list)

        time_id = int(temp[1].strip())
        num = int(temp[2].strip())
        timeseries[group_id][time_id] = num

    return (timeseries, max_t)


#num_bitmasks = 3
#bit = 20
def get_starting_initial_nodes_single_location(induced_graph, timeseries_list, T):

    #get all distances
    #M: key: node; value: [ 1[bits], 2,...,h, ..T ]
    M = {}
    for node in induced_graph.nodes():
        M[node] = []
        for h in xrange(T):
            M[node].append([])

            for i in xrange(num_bitmasks):
                num = 0
                if h == 0:
                    prob = 0.5
                    for i in xrange(bit-1, -1, -1):
                        if random.random() < prob:
                            num += (1<< i)
                            break

                        prob = prob*0.5

                M[node][h].append(num)

    #M[node][h][j]: a bit
    for i in xrange(1,T):
        for node in induced_graph.nodes():
            for j in xrange(num_bitmasks):
                M[node][i][j] = M[node][i-1][j]

        #update edges
        for e in induced_graph.edges_iter():
            (n1, n2) = e
            for j in xrange(num_bitmasks):
                M[n1][i][j]  = (M[n1][i][j] ^ M[n2][i-1][j])
                M[n2][i][j]  = (M[n2][i][j] ^ M[n1][i-1][j])


    S = []
    is_in = {}
    K = timeseries_list[0]
    M_S = [ [0 for j in xrange(num_bitmasks)] for i in xrange(T) ]
    #print len(M_S)
    #print M_S, M_S[h][num_bitmasks]

    for k in xrange(K):
        max_node, max_delta = None, -1
        for node in induced_graph.nodes():
            if node not in is_in:
                delta_S = 0
                CN = 0
                CF_S_union_node = 0
                CF_S = 0
                for h in xrange(T):
                    CN += timeseries_list[h]
                    CF_S += convert_to_neighborhood(M_S[h], num_bitmasks, bit)
                    tmp = [ ( M_S[h][j] ^ M[node][h][j] )   for j in xrange(num_bitmasks)]
                    CF_S_union_node += convert_to_neighborhood(tmp, num_bitmasks, bit)

                    theta_S = 0
                    if CF_S < CN:
                        theta_S = CN - CF_S
                    theat_S_union_node = 0
                    if CF_S_union_node < CN:
                        theat_S_union_node = CN - CF_S_union_node

                    delta_S += (theta_S - theat_S_union_node)
                if delta_S > max_delta:
                    max_node = node
                    max_delta = delta_S
        if max_node:
            S.append(max_node)
            is_in[max_node] = True

        for h in xrange(T):
            for j in xrange(num_bitmasks):
                if max_node in M:
                    M_S[h][j] = ( M_S[h][j] ^ M[max_node][h][j])


    #print S
    return S

#input: a list [ 000, 100, 101, 010 ]
def convert_to_neighborhood(M_S_h_list, num_bitmasks, bit):
    b = 0
    for num in M_S_h_list:
        for i in xrange((bit)-1):
            shift = bit -1-i
            mask = ( 1<< (shift-1) )
            t = mask & num
            if t ==0:
                b += i
                break

    #In = 2^b / 0.77351
    b = b/len(M_S_h_list)
    return 2**b/0.77351


def get_starting_initial_nodes(G, group_dict, invert_group_dict, timeseries_dict, T):
    S = []

    for loc in timeseries_dict:
        if loc not in group_dict:
            continue
        induced_graph =  G.subgraph(group_dict[loc])
        #print loc, induced_graph.nodes(), induced_graph.edges()
        #print loc
        S += get_starting_initial_nodes_single_location(induced_graph, timeseries_dict[loc], T)

    return S


def get_mapping_multi_locations(G, group_dict, invert_group_dict, timeseries_dict, T):

    #get the starting infected node S
    S = get_starting_initial_nodes(G, group_dict, invert_group_dict, timeseries_dict, T)
    #print S

    #get cascades
    #Layer: [  {loc1: [a,b,c], loc2: [c,d,e]}, {}, {}, {}  ]
    Layer = []
    layered_graph = nx.DiGraph()

    Layer.append({})
    for node in S:
        layered_graph.add_node(node)
        #print node
        group = invert_group_dict[node]
        if group not in Layer[-1]:
            Layer[-1][group] =[]

        Layer[-1][group].append(node)

    q = []
    q += S
    is_visit ={}
    for node in S:
        is_visit[node]=0

    i = 1
    while q:
        tmp = []
        for node in q:
            for nei in G.neighbors(node):
                if (nei in is_visit) and (is_visit[nei]==i):
                    layered_graph.add_edge(node,nei)

                if nei not in is_visit:
                    layered_graph.add_edge(node, nei)
                    tmp.append(nei)
                    is_visit[nei] = i

        Layer.append({})
        for node in tmp:
            group = invert_group_dict[node]
            if group not in Layer[-1]:
                Layer[-1][group] =[]

            Layer[-1][group].append(node)

        q = tmp
        i += 1
        if i>=T-1:
            break

    #get cascade
    cascades = []
    for i in xrange(number_of_cascade):

        induced_nodes = []

        #timeseries[group_id][time_id] = num
        for group_id in timeseries_dict:
            tmp=set([])
            for i in xrange(len(timeseries_dict[group_id])):
                CN = timeseries_dict[group_id][i]
                #Layer: [  {loc1: [a,b,c], loc2: [c,d,e]}, {}, {}, {}  ]
                if i<len(Layer) and group_id in Layer[i]:
                    tmp = tmp.union(set(Layer[i][group_id]))
                #print group_id, i
                #print CN
                #print tmp
                try:
                    select = random.sample( tmp, CN)
                except:

                    continue
                induced_nodes += select
                tmp = tmp - set(select)
        induced_graph =  layered_graph.subgraph(induced_nodes)
        cascades.append(induced_graph)
    return [cascades, S]

#credit_dict = {node: credit}
def compute_credit(group_dict, cascades):
    credit_dict = {}
    for loc in group_dict:
        for node in group_dict[loc]:
            credit_dict[node] = 0
    for cascade in cascades:
        c ={}
        for node in cascade.nodes():
            c[node] = 1
        #print cascade.edges()
        node_order=nx.topological_sort(cascade)
        node_order.reverse()
        #print node_order

        #c(u) = 1+ sum_u's out-nei (c(u))
        for node in node_order:
            for outnei in cascade.successors(node):
                c[node] += c[outnei]

        for node in cascade.nodes():
            credit_dict[node] += c[node]


    return credit_dict

#G: graph
#group_dict: loc:[1, 2,3]
#cascade: directed graph (DAG)
#S: initial starting point
#k: budget
def group_immunization(G, group_dict, cascades, S, k, edgelist_file):
    vacc_alloc = {}
    for loc in group_dict:
        vacc_alloc[loc] = 0

    #ran = int(time.time())
    #result_file = edgelist_file + "-result-"+str(ran)
    result_file = edgelist_file + "-result"
    f = open(result_file, "w")
    for m in xrange(k):
        best_loc = None
        best_credit = -1
        credit_dict = compute_credit(group_dict, cascades)
        for loc in group_dict:
            credit_sum = 0

            for node in group_dict[loc]:
                credit_sum += credit_dict[node]
                #credit_dict = {node: credit}
            credit_sum = credit_sum*1.0/len(group_dict[loc])
            if credit_sum > best_credit:
                best_credit = credit_sum
                best_loc = loc

        #assign vaccines to the best location
        #print best_loc, m
        f.write("%d, %s\n" % (m, best_loc))
        vacc_alloc[best_loc] += 1

        #randomly sample a node from the best location
        node = random.choice(group_dict[best_loc])
        for cascade in cascades:
            if node in cascade:
                cascade.remove_node(node)

    f.close()
    return vacc_alloc


#python data-immu.py edgelist groups infection_num budget
#@parameters
#edgeedge: 1 2
#groups: loc1, 1, 2, 3
#infection_timeseries: loc1, t1, 1
#@return: vaccine allocation vector: loc1: #vaccines
def main():
    start = time.time()
    start1 = time.time()
    edgelist_file = sys.argv[1]
    groups_file= sys.argv[2]
    infection_num_file = sys.argv[3]
    k = int(sys.argv[4])

    #group_dict: key: group_id; value: node_list
    #timeseries: timeseries[group_id][timestep]
    G = create_graph(edgelist_file)
    (group_dict, invert_group_dict) = get_groups(groups_file)
    (timeseries_dict, T) = get_timeseries(infection_num_file)

    print "================="
    #print "nodes:", G.nodes()
    #print "groups:", group_dict
    #print "invert_group_dict", invert_group_dict
    #print "timeseries:", timeseries_dict
    #print "max_timestep", T
    #print "=================\n"

    #algorithm 6
    #return mapping_dict. key: loc_id; value: node_lists.
    [cascades, S] = get_mapping_multi_locations(G, group_dict, invert_group_dict, timeseries_dict, T)
    #print "result: ", cascades[0].nodes(), cascades[1].nodes()
    #print "initial infected S: ", S
    end = time.time()
    print "sample cascade running time:", end-start

    #print "immunization"
    start = time.time()
    vacc_alloc = group_immunization(G, group_dict, cascades, S, k, edgelist_file)
    #print vacc_alloc
    end = time.time()
    print "immunization running time:", end- start

    print "total running time", end-start1

if __name__ == "__main__":
    main()
